/********************************************************************
*               MINI-COMPILADOR PARA O MINI-PASCAL                  *
* ----------------------------------------------------------------- *
*      Membros: Anderson de Rezende Rocha - DCC/UFLA                *
*               undersun@comp.ufla.br                              *
*                                                                   *
*               Jlio Csar Alves - DCC/UFLA                        *
*               jcalves@comp.ufla.br                                *
* ----------------------------------------------------------------- *
*          COMPILADORES: Prof. Mrio Luis R. Oliveira               *
* ----------------------------------------------------------------- *
* Desenvolvida inicialmente para a disciplina de Linguagens Formais *
*         e autmatos. Prof. Jones Oliveira Albuquerque.            *
*                                                                   *
*  Anderson de Rezende Rocha, Flvio Luiz Alves, Jlio Csar Alves  *
*                       Rodrigo Nazar Leite                        *
********************************************************************/

/*
    OBSERVAES:
        + Esta classe define um automato finito deterministico qualquer, com mensagem de saida no estado 
          (Maquina de Moore) e na ligao (Mquina de Mealy)
        + Ela eh template em relacao ao alfabeto, ou seja, o alfabeto
          pode ser qualquer classe definida pelo usuario desta classe
        + Depois de inserir todos os estados e producoes se for utilizada a funcao ProcessarSimbolo()
          o usurio eh responsavel por Reinicializar o Automato antes de passar o 1o. simbolo
        + Esta implementacao utiliza a ideia de lista de listas

    MTODOS:
        + clAutomato<TSimbolo>(int idEstadoInicial, bool final, char saida[36]);
        + ~clAutomato<TSimbolo>();
        + bool InserirEstado(int idEstado, bool final, char saida[36]);
        + bool InserirProducao(int idEstadoPartida, int idEstadoDestino, TSimbolo* simbolo);
        + bool InserirProducao(int idEstadoPartida, int idEstadoDestino, TSimbolo* simbolo, string saida);
        + bool ProcessarPalavra(clLista<TSimbolo*> palavra);
        + bool ProcessarSimbolo(TSimbolo *simbolo, char saida[36], bool ehUltimoSimbolo = false);
        + bool TentarProcessarSimbolo(TSimbolo *simbolo, char saidaDoEstado[36],
                                    string &saidaDaLigacao, bool ehUltimoSimbolo = false);
        + bool GetPalavraFoiAceita();
        + bool GetEstadoAtualEhFinal();
        + void ReinicializarAutomato();
        + clAutomato<TSimbolo> *CopiarAutomato();

        - void Imprimir();
        - clEstado<TSimbolo>* EncontrarEstado(int id);
*/


#ifndef AUTOMATO_H
#define AUTOMATO_H

#include "estado.h"
#include <string>

using namespace std;

// Obs: noh e estado so a mesma coisa

// Definio da classe clAutomato como template
template <class TSimbolo>
// A classe TSimbolo deve definir o operador ==
class clAutomato
{
    public:
        // Construtor que recebe um identificador para o noh, um booleano [verdadeiro/falso]
        // identificando se este e final ou nao e um char que ser a mensagem de saida do primeiro
        // noh do automato. Esta mensagem fica no noh pois usamos a abordagem das maquinas
        // de moore.
        clAutomato<TSimbolo>(int idEstadoInicial, bool final, char saida[36]);
        // Destrutor padrao para desalocar memria
        ~clAutomato<TSimbolo>();
        // Insere um novo estado no automato, recebe um identificador numrico, um booleano
        // dizendo se o noh  final e a saida que este estado ter
        bool InserirEstado(int idEstado, bool final, char saida[36]);
        // Insere uma producao a partir de um estado de inicio e um estado de destino, mais
        // o simbolo especifico para esta producao
        bool InserirProducao(int idEstadoPartida, int idEstadoDestino, TSimbolo* simbolo);
        // Insere uma producao a partir de um estado de inicio e um estado de destino, mais
        // o simbolo especifico para esta producao e a saida. Obs: Pode parecer estranho a saida do estado
        // ser um char[36] e a saida da ligacao ser um string. Mas eh que originalmente o automato
        // tinha saida apenas no estado. Preferiu-se utilizar agora string, mas mantendo a compatibilidade
        bool InserirProducao(int idEstadoPartida, int idEstadoDestino, TSimbolo* simbolo, string saida);
        // Processa uma palavra inteira de entrada de uma vez
        bool ProcessarPalavra(clLista<TSimbolo*> palavra);
        // Processa um simbolo da entrada de cada vez
        // o boolean do parametro serve para informar ao automato que a fita acabou
        bool ProcessarSimbolo(TSimbolo *simbolo, char saida[36], bool ehUltimoSimbolo = false);
        // Processa um simbolo da entrada de cada vez
        // o boolean do parametro serve para informar ao automato que a fita acabou
        // A funcao abaixo quando nao ha producao para um simbolo, retorna false, mas deixa o estadoAtual
        // como se nao tivesse havido a producao
        bool TentarProcessarSimbolo(TSimbolo *simbolo, char saidaDoEstado[36],
                                    string &saidaDaLigacao, bool ehUltimoSimbolo = false);
        // Retorna se a palavra foi ou nao aceita
        bool GetPalavraFoiAceita();
        // Retorna se a o estado atual eh um estado final
        bool GetEstadoAtualEhFinal();
        // Reinicializa o automato para processamento de uma nova palavra
        void ReinicializarAutomato();
        // Devolve uma referencia para uma copia do automato
        clAutomato<TSimbolo> *CopiarAutomato();

        void Imprimir();

    private:
        // Encontra um estado especifico para, por exemplo, ser usado pela
        // funcao que insere uma producao
        clEstado<TSimbolo>* EncontrarEstado(int id);
        // Guarda qual o estado inicial do automato
        clEstado<TSimbolo>* estadoInicial;
        // Estado atual de processamento de uma palavra em especifico
        clEstado<TSimbolo>* estadoAtual;
        // Guarda se a palavra esta sendo ou nao aceita ate o momento
        bool palavraFoiAceita;
};

// Construtor padrao
template <class TSimbolo>
clAutomato<TSimbolo>::clAutomato<TSimbolo>(int idEstadoInicial, bool final, char saida[36])
{
    // aloca o estado inicial
    estadoInicial = new clEstado<TSimbolo>(idEstadoInicial, final, saida);
    // para evitar erros
    estadoAtual = estadoInicial;
}

// Destrutor, enquanto houver membros alocados dinamicamente desaloque-os
template <class TSimbolo>
clAutomato<TSimbolo>::~clAutomato<TSimbolo>()
{
    clEstado<TSimbolo> *estado;
    estado = estadoInicial;
    while (estadoAtual != NULL)
    {
        estado = estadoAtual;
        estadoAtual = estado->getProximoEstado();
        delete estado;
    }
}

// Insere um estado, a partir do idEstado. Se este ja existir o identificador nao
// sera inserido, caso contrario, insira-o no automato.
template <class TSimbolo>
bool clAutomato<TSimbolo>::InserirEstado(int idEstado, bool final, char saida[36])
{
    if (EncontrarEstado(idEstado))
        return false;
    else
    {
        clEstado<TSimbolo> *estado;
        estado = new clEstado<TSimbolo>(idEstado, final, saida);
        estadoAtual->setProximoEstado(estado);
        estadoAtual = estado;
    }
    return true;
}

// Insere uma producao. Busca se o estado que contem o id de partida e o estado
// que contem o id de destino existem. Retorne-os. Se forem diferentes de NULL entao
// ligue-os insira uma producao com o simbolo passado, entre eles.
template <class TSimbolo>
bool clAutomato<TSimbolo>::InserirProducao(int idEstadoPartida, int idEstadoDestino,
                                                                    TSimbolo* simbolo)
{
    return (InserirProducao(idEstadoPartida, idEstadoDestino, simbolo, ""));
}

// Insere uma producao. Busca se o estado que contem o id de partida e o estado
// que contem o id de destino. Retorne-os. Se forem diferentes de NULL entao
// ligue-os insira uma producao com o simbolo e a saida passados, entre eles.
template <class TSimbolo>
bool clAutomato<TSimbolo>::InserirProducao(int idEstadoPartida, int idEstadoDestino,
                                            TSimbolo* simbolo, string saida)
{
    clEstado<TSimbolo> *estadoPartida = EncontrarEstado(idEstadoPartida);
    clEstado<TSimbolo> *estadoChegada = EncontrarEstado(idEstadoDestino);
    if (estadoPartida == NULL || estadoChegada == NULL)
        return false;
    else
        return (estadoPartida->InserirProducao(simbolo, estadoChegada, saida));
}

// Retorna se uma palavra foi aceita ou nao
template <class TSimbolo>
bool clAutomato<TSimbolo>::GetPalavraFoiAceita()
{
    return palavraFoiAceita;
}

// Retorna se uma palavra foi aceita ou nao
template <class TSimbolo>
bool clAutomato<TSimbolo>::GetEstadoAtualEhFinal()
{
    return (estadoAtual->getEhFinal());
}

// Reinicializa o automato para processamento de outra palavra qualquer
template <class TSimbolo>
void clAutomato<TSimbolo>::ReinicializarAutomato()
{
    estadoAtual = estadoInicial;
    palavraFoiAceita = false;
}

// Funcao delta-estendida, processa uma palavra inteira de uma vez
// Por definio apenas aplica a funcao delta comun "ProcessarSimbolo"
// varias vezes. Na verdade o numero de vezes eh a cardinalidade da palavra
// de entrada.
template <class TSimbolo>
bool clAutomato<TSimbolo>::ProcessarPalavra(clLista<TSimbolo*> palavra)
{
    ReinicializarAutomato();
    clNohLista<TSimbolo*> *simbolo;
    simbolo = palavra.Cabeca();
    while(estadoAtual != NULL && simbolo != NULL)
    {
        ProcessarSimbolo(simbolo->TValor());
        simbolo = simbolo->Proximo();
    }

    if (estadoAtual != NULL)
        palavraFoiAceita = estadoAtual->getEhFinal();
    else
        palavraFoiAceita = false;
    return (getPalavraFoiAceita());
}

// O usuario deve reinicializar o automato quando necessario
// Retorna verdadeiro se ha producao para tal simbolo
// A saida eh retornada por dentro
template <class TSimbolo>
bool clAutomato<TSimbolo>::ProcessarSimbolo(TSimbolo *simbolo, char saida[36],
                                                              bool ehUltimoSimbolo)
{
    estadoAtual = estadoAtual->Consumir(simbolo);
    if (ehUltimoSimbolo && estadoAtual != NULL && estadoAtual->getEhFinal())
    {
        palavraFoiAceita = true;
        strcpy(saida,estadoAtual->getSaida());
        return GetPalavraFoiAceita();
    }
    else if (estadoAtual != NULL)
    {
        strcpy(saida,estadoAtual->getSaida());
        return true;
    }
    // cout << "fim da Processar Simbolo" << endl;
    return false;
}

// A funcao abaixo quando nao ha producao para um simbolo, retorna false, mas deixa o estadoAtual
// como se nao tivesse havido a producao
// O usuario deve reinicializar o automato quando necessario
// Retorna verdadeiro se ha producao para tal simbolo
// A saidas sao retornadas por dentro
template <class TSimbolo>
bool clAutomato<TSimbolo>::TentarProcessarSimbolo(TSimbolo *simbolo, char saidaDoEstado[36],
                                                  string &saidaDaLigacao, bool ehUltimoSimbolo)
{
    int debugId = estadoAtual->getId();
    clEstado<TSimbolo>* estadoAnterior;
    estadoAnterior = estadoAtual;
    estadoAtual = estadoAtual->Consumir(simbolo, saidaDaLigacao);
    if (ehUltimoSimbolo && estadoAtual != NULL && estadoAtual->getEhFinal())
    {
        palavraFoiAceita = true;
        strcpy(saidaDoEstado,estadoAtual->getSaida());
        return GetPalavraFoiAceita();
    }
    else if (estadoAtual != NULL)
    {
        strcpy(saidaDoEstado,estadoAtual->getSaida());
        return true;
    }
    else
    {
        estadoAtual = estadoAnterior;
        return false;
    }
}

// Encontra um estado em especial a partir de seu identificador
// Retorna NULL se o estado no foi encontrado
template <class TSimbolo>
clEstado<TSimbolo>*  clAutomato<TSimbolo>::EncontrarEstado(int id)
{
    clEstado<TSimbolo>* estado;
    estado = estadoInicial;
    while (estado != NULL)
    {
        if (id == estado->getId())
            return estado;
        else
            estado = estado->getProximoEstado();
    }
    return estado;
}

// Devolve uma referencia para uma copia do automato
template <class TSimbolo>
clAutomato<TSimbolo> *clAutomato<TSimbolo>::CopiarAutomato()
{
    clAutomato<TSimbolo> *resposta;
    clEstado<TSimbolo> *estado;

    resposta = new clAutomato<TSimbolo>(estadoInicial->getId(), estadoInicial->getEhFinal(),
                                        estadoInicial->getSaida());
    // Copiando os estados
    estado = estadoInicial->getProximoEstado();
    while (estado != NULL)
    {
        resposta->InserirEstado(estado->getId(), estado->getEhFinal(), estado->getSaida());
        estado = estado->getProximoEstado();
    }

    // Copiando as producoes
    estado = estadoInicial;
    while (estado != NULL)
    {
        clNohLista<clLigacao<TSimbolo> > *noh;
        noh = estado->getProducoes()->Primeiro();
        while (noh != NULL)
        {
            resposta->InserirProducao(estado->getId(), noh->Valor().getDestino()->getId(),
                                     noh->Valor().getSimbolo(), noh->Valor().getSaida());
            noh = noh->Proximo();
        }
        estado = estado->getProximoEstado();
    }
    return resposta;
}

// Imprime o o estado interno do autmato
template <class TSimbolo>
void clAutomato<TSimbolo>::Imprimir()
{
    clEstado<TSimbolo> *estado;
    clNohLista<clLigacao<TSimbolo> > *noh;
    estado = estadoInicial;

    while (estado != NULL)
    {
        cout << estado->getId() << " - ";
        noh = estado->getProducoes()->Primeiro();
        while (noh != NULL)
        {
            cout << "(" << noh->Valor().getDestino()->getId() << "," << *(noh->Valor().getSimbolo()) << ","
            << noh->Valor().getSaida() << ") ";
            noh = noh->Proximo();
        }
        cout << endl;
        estado = estado->getProximoEstado();
    }
}

#endif


